@@ -115,6 +115,9 @@ data S3Error
       , s3ErrorHostId :: Maybe T.Text -- Error/HostId
       , s3ErrorAccessKeyId :: Maybe T.Text -- Error/AWSAccessKeyId
       , s3ErrorStringToSign :: Maybe B.ByteString -- Error/StringToSignBytes (hexadecimal encoding)
+      , s3ErrorBucket :: Maybe T.Text -- Error/Bucket
+      , s3ErrorEndpointRaw :: Maybe T.Text -- Error/Endpoint (i.e. correct bucket location)
+      , s3ErrorEndpoint :: Maybe B.ByteString -- Error/Endpoint without the bucket prefix
       }
     deriving (Show, Typeable)
 
@@ -245,7 +248,7 @@ s3BinaryResponseConsumer inner metadata resp = do
       let m = S3Metadata { s3MAmzId2 = amzId2, s3MRequestId = requestId }
       liftIO $ tellMetadataRef metadata m
 
-      if HTTP.responseStatus resp >= HTTP.status400
+      if HTTP.responseStatus resp >= HTTP.status300
         then s3ErrorResponseConsumer resp
         else inner resp
 
@@ -269,6 +272,9 @@ s3ErrorResponseConsumer resp
                            let resource = listToMaybe $ root $/ elContent "Resource"
                                hostId = listToMaybe $ root $/ elContent "HostId"
                                accessKeyId = listToMaybe $ root $/ elContent "AWSAccessKeyId"
+                               bucket = listToMaybe $ root $/ elContent "Bucket"
+                               endpointRaw = listToMaybe $ root $/ elContent "Endpoint"
+                               endpoint = T.encodeUtf8 <$> (T.stripPrefix (fromMaybe "" bucket <> ".") =<< endpointRaw)
                                stringToSign = do unprocessed <- listToMaybe $ root $/ elCont "StringToSignBytes"
                                                  bytes <- mapM readHex2 $ words unprocessed
                                                  return $ B.pack bytes
@@ -280,6 +286,9 @@ s3ErrorResponseConsumer resp
                                       , s3ErrorHostId = hostId
                                       , s3ErrorAccessKeyId = accessKeyId
                                       , s3ErrorStringToSign = stringToSign
+                                      , s3ErrorBucket = bucket
+                                      , s3ErrorEndpointRaw = endpointRaw
+                                      , s3ErrorEndpoint = endpoint
                                       }
 
 type CanonicalUserId = T.Text
@@ -317,20 +326,25 @@ writeCannedAcl AclLogDeliveryWrite       = "log-delivery-write"
 
 data StorageClass
     = Standard
+    | StandardInfrequentAccess
     | ReducedRedundancy
     | Glacier
+    | OtherStorageClass T.Text
     deriving (Show)
 
-parseStorageClass :: MonadThrow m => T.Text -> m StorageClass
-parseStorageClass "STANDARD"           = return Standard
-parseStorageClass "REDUCED_REDUNDANCY" = return ReducedRedundancy
-parseStorageClass "GLACIER"            = return Glacier
-parseStorageClass s = throwM . XmlException $ "Invalid Storage Class: " ++ T.unpack s
+parseStorageClass :: T.Text -> StorageClass
+parseStorageClass "STANDARD"           = Standard
+parseStorageClass "STANDARD_IA"        = StandardInfrequentAccess
+parseStorageClass "REDUCED_REDUNDANCY" = ReducedRedundancy
+parseStorageClass "GLACIER"            = Glacier
+parseStorageClass s                    = OtherStorageClass s
 
 writeStorageClass :: StorageClass -> T.Text
-writeStorageClass Standard          = "STANDARD"
-writeStorageClass ReducedRedundancy = "REDUCED_REDUNDANCY"
-writeStorageClass Glacier           = "GLACIER"
+writeStorageClass Standard                 = "STANDARD"
+writeStorageClass StandardInfrequentAccess = "STANDARD_IA"
+writeStorageClass ReducedRedundancy        = "REDUCED_REDUNDANCY"
+writeStorageClass Glacier                  = "GLACIER"
+writeStorageClass (OtherStorageClass s) = s
 
 data ServerSideEncryption
     = AES256
@@ -383,7 +397,7 @@ parseObjectInfo el
          lastModified <- forceM "Missing object LastModified" $ el $/ elContent "LastModified" &| time
          eTag <- force "Missing object ETag" $ el $/ elContent "ETag"
          size <- forceM "Missing object Size" $ el $/ elContent "Size" &| textReadInt
-         storageClass <- forceM "Missing object StorageClass" $ el $/ elContent "StorageClass" &| parseStorageClass
+         storageClass <- forceM "Missing object StorageClass" $ el $/ elContent "StorageClass" &| return . parseStorageClass
          owner <- case el $/ Cu.laxElement "Owner" &| parseUserInfo of
                     (x:_) -> fmap' Just x
                     [] -> return Nothing
